@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2012 Tim Ruehsen
- * Copyright (c) 2016-2021 Free Software Foundation, Inc.
+ * Copyright (c) 2016-2024 Free Software Foundation, Inc.
  *
  * This file is part of libwget.
  *
@@ -53,19 +53,20 @@ static struct random_data state;
 static wget_thread_mutex mutex;
 static bool initialized;
 
-static void __attribute__ ((constructor)) random_init(void)
-{
-	if (!initialized) {
-		wget_thread_mutex_init(&mutex);
-		initialized = 1;
-	}
-}
-
-static void __attribute__ ((destructor)) random_exit(void)
+static void  random_exit(void)
 {
 	if (initialized) {
 		wget_thread_mutex_destroy(&mutex);
-		initialized = 0;
+		initialized = false;
+	}
+}
+
+INITIALIZER(random_init)
+{
+	if (!initialized) {
+		wget_thread_mutex_init(&mutex);
+		initialized = true;
+		atexit(random_exit);
 	}
 }
 
