@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2012-2015 Tim Ruehsen
- * Copyright (c) 2015-2022 Free Software Foundation, Inc.
+ * Copyright (c) 2015-2024 Free Software Foundation, Inc.
  *
  * This file is part of libwget.
  *
@@ -52,6 +52,9 @@
 #ifdef WITH_OCSP
 #	include <gnutls/ocsp.h>
 #endif
+#ifdef WITH_LIBDANE
+#	include <gnutls/dane.h>
+#endif
 #include <gnutls/crypto.h>
 #include <gnutls/abstract.h>
 
@@ -100,17 +103,20 @@ static struct config {
 		key_type;
 	bool
 		check_certificate : 1,
+		report_invalid_cert : 1,
 		check_hostname : 1,
 		print_info : 1,
 		ocsp : 1,
 		ocsp_date : 1,
 		ocsp_stapling : 1,
-		ocsp_nonce : 1;
+		ocsp_nonce : 1,
+		dane : 1;
 } config = {
 	.check_certificate = 1,
+	.report_invalid_cert = 1,
 	.check_hostname = 1,
 #ifdef WITH_OCSP
-	.ocsp = 1,
+	.ocsp = 0,
 	.ocsp_stapling = 1,
 #endif
 	.ca_type = WGET_SSL_X509_FMT_PEM,
@@ -118,6 +124,7 @@ static struct config {
 	.key_type = WGET_SSL_X509_FMT_PEM,
 	.secure_protocol = "AUTO",
 	.ca_directory = "system",
+	.ca_file = "system",
 #ifdef WITH_LIBNGHTTP2
 	.alpn = "h2,http/1.1",
 #endif
@@ -128,6 +135,8 @@ struct session_context {
 		hostname;
 	wget_hpkp_stats_result
 		stats_hpkp;
+	uint16_t
+		port;
 	bool
 		ocsp_stapling : 1,
 		valid : 1,
@@ -139,6 +148,8 @@ static gnutls_certificate_credentials_t
 static gnutls_priority_t
 	priority_cache;
 
+#define error_printf_check(...) if (config.report_invalid_cert) wget_error_printf(__VA_ARGS__)
+
 /**
  * \param[in] key An identifier for the config parameter (starting with `WGET_SSL_`) to set
  * \param[in] value The value for the config parameter (a NULL-terminated string)
@@ -251,6 +262,7 @@ void wget_ssl_set_config_object(int key, void *value)
  * These are the parameters that can be set (\p key can have any of these values):
  *
  *  - WGET_SSL_CHECK_CERTIFICATE: whether certificates should be verified (1) or not (0)
+ *  - WGET_SSL_REPORT_INVALID_CERT: whether to print (1) errors/warnings regarding certificate verification or not (0)
  *  - WGET_SSL_CHECK_HOSTNAME: whether or not to check if the certificate's subject field
  *  matches the peer's hostname. This check is done according to the rules in [RFC 6125](https://tools.ietf.org/html/rfc6125)
  *  and typically involves checking whether the hostname and the common name (CN) field of the subject match.
@@ -282,9 +294,11 @@ void wget_ssl_set_config_int(int key, int value)
 {
 	switch (key) {
 	case WGET_SSL_CHECK_CERTIFICATE: config.check_certificate = (char)value; break;
+	case WGET_SSL_REPORT_INVALID_CERT: config.report_invalid_cert = (char)value; break;
 	case WGET_SSL_CHECK_HOSTNAME: config.check_hostname = (char)value; break;
 	case WGET_SSL_CA_TYPE: config.ca_type = (char)value; break;
 	case WGET_SSL_CERT_TYPE: config.cert_type = (char)value; break;
+	case WGET_SSL_DANE: config.dane = (char)value; break;
 	case WGET_SSL_KEY_TYPE: config.key_type = (char)value; break;
 	case WGET_SSL_PRINT_INFO: config.print_info = (char)value; break;
 	case WGET_SSL_OCSP: config.ocsp = (char)value; break;
@@ -312,7 +326,7 @@ static void print_x509_certificate_info(gnutls_session_t session)
 {
 	const char *name;
 	char dn[128], timebuf[64];
-	unsigned char digest[20];
+	unsigned char digest[64];
 	unsigned char serial[40];
 	size_t dn_size = sizeof(dn);
 	size_t digest_size = sizeof (digest);
@@ -347,7 +361,7 @@ static void print_x509_certificate_info(gnutls_session_t session)
 			info_printf(_("  Expires: %s"), safe_ctime(expired, timebuf, sizeof(timebuf)));
 
 			if (!gnutls_fingerprint(GNUTLS_DIG_MD5, &cert_list[ncert], digest, &digest_size)) {
-				char digest_hex[digest_size * 2 + 1];
+				char digest_hex[sizeof(digest) * 2 + 1];
 
 				wget_memtohex(digest, digest_size, digest_hex, sizeof(digest_hex));
 
@@ -355,9 +369,9 @@ static void print_x509_certificate_info(gnutls_session_t session)
 			}
 
 			if (!gnutls_x509_crt_get_serial(cert, serial, &serial_size)) {
-				char serial_hex[digest_size * 2 + 1];
+				char serial_hex[sizeof(serial) * 2 + 1];
 
-				wget_memtohex(digest, digest_size, serial_hex, sizeof(serial_hex));
+				wget_memtohex(serial, serial_size, serial_hex, sizeof(serial_hex));
 
 				info_printf(_("  Serial number: %s\n"), serial_hex);
 			}
@@ -689,7 +703,7 @@ static int check_ocsp_response(gnutls_x509_crt_t cert,
 	gnutls_ocsp_resp_t resp;
 	int ret = -1, rc;
 	unsigned int status, cert_status;
-	time_t rtime, vtime, ntime, now;
+	time_t rtime = 0, vtime = 0, ntime = 0, now;
 	char timebuf[64];
 
 	now = time(NULL);
@@ -821,7 +835,7 @@ static void add_cert_to_ocsp_cache(gnutls_x509_crt_t cert, bool valid)
 }
 
 /* OCSP check for the peer's certificate. Should be called
- * only after the certificate list verication is complete.
+ * only after the certificate list verification is complete.
  * Returns:
  * 0: certificate is revoked
  * 1: certificate is ok
@@ -846,6 +860,11 @@ static int cert_verify_ocsp(gnutls_x509_crt_t cert, gnutls_x509_crt_t issuer)
 		return -1;
 	}
 
+	if (!resp) {
+		debug_printf("Missing response from OCSP server\n");
+		return -1;
+	}
+
 	/* verify and check the response for revoked cert */
 	ret = check_ocsp_response(cert, issuer, resp, &nonce);
 	wget_buffer_free(&resp);
@@ -923,6 +942,17 @@ out:
 	return ret; // Pubkey not found
 }
 
+static void print_verification_status(gnutls_session_t session, const char *tag, int status) {
+	gnutls_datum_t out;
+
+	if (gnutls_certificate_verification_status_print(
+		status, gnutls_certificate_type_get(session), &out, 0) == GNUTLS_E_SUCCESS)
+	{
+		error_printf_check("%s: %s\n", tag, out.data); // no translation
+		xfree(out.data);
+	}
+}
+
 /* This function will verify the peer's certificate, and check
  * if the hostname matches, as well as the activation, expiration dates.
  */
@@ -933,15 +963,15 @@ static int verify_certificate_callback(gnutls_session_t session)
 	unsigned int cert_list_size;
 	int ret = -1, err, ocsp_ok = 0, pinning_ok = 0;
 	gnutls_x509_crt_t cert = NULL, issuer = NULL;
-	const char *hostname;
 	const char *tag = config.check_certificate ? _("ERROR") : _("WARNING");
 #ifdef WITH_OCSP
+	bool skip_server_cert_check = false;
 	unsigned nvalid = 0, nrevoked = 0, nignored = 0;
 #endif
 
 	// read hostname
 	struct session_context *ctx = gnutls_session_get_ptr(session);
-	hostname = ctx->hostname;
+	const char *hostname = ctx->hostname;
 
 	/* This verification function uses the trusted CAs in the credentials
 	 * structure. So you must have installed one or more CA certificates.
@@ -953,7 +983,7 @@ static int verify_certificate_callback(gnutls_session_t session)
 #endif
 //		if (wget_get_logger(WGET_LOGGER_DEBUG))
 //			_print_info(session);
-		error_printf(_("%s: Certificate verification error\n"), tag);
+		error_printf_check(_("%s: Certificate verification error\n"), tag);
 		goto out;
 	}
 
@@ -978,39 +1008,70 @@ static int verify_certificate_callback(gnutls_session_t session)
 #endif
 
 #if GNUTLS_VERSION_NUMBER >= 0x030104
+#ifdef WITH_LIBDANE
+	// If CA cert verification failed due to missing certificates, we try DANE verification (if requested by the user).
 	if (status) {
-		gnutls_datum_t out;
-
-		if (gnutls_certificate_verification_status_print(
-			status, gnutls_certificate_type_get(session), &out, 0) == GNUTLS_E_SUCCESS)
-		{
-			error_printf("%s: %s\n", tag, out.data); // no translation
-			xfree(out.data);
+		if (!config.dane) {
+			print_verification_status(session, tag, status);
+			goto out;
+		}
+		if (status != (GNUTLS_CERT_INVALID | GNUTLS_CERT_SIGNER_NOT_FOUND)) {
+			print_verification_status(session, tag, status);
+			goto out;
 		}
 
-		goto out;
+		// GNUTLS_CERT_SIGNER_NOT_FOUND indicates that no matching CA cert exists.
+
+		unsigned verify = 0;
+
+		int rc = dane_verify_session_crt(NULL, session, hostname, "tcp", ctx->port, 0,
+			DANE_VFLAG_FAIL_IF_NOT_CHECKED,
+			&verify);
+
+		if (rc < 0) {
+			debug_printf("DANE verification error for %s: %s\n", hostname, dane_strerror(rc));
+			goto out;
+		} else if (verify) {
+			gnutls_datum_t out;
+			rc = dane_verification_status_print(verify, &out, 0);
+			if (rc < 0) {
+				error_printf(_("DANE verification print error for %s: %s\n"), hostname, dane_strerror(rc));
+			} else {
+				error_printf(_("DANE verification failed for %s: %s\n"), hostname, out.data);
+			}
+			gnutls_free(out.data);
+			goto out;
+		} else {
+			debug_printf("DANE verification: %s\n", dane_strerror(rc));
+		}
 	}
+#else
+	if (status) {
+		print_verification_status(session, tag, status);
+		goto out;
+	}
+#endif
 #else
 	if (status) {
 		if (status & GNUTLS_CERT_INVALID)
-			error_printf(_("%s: The certificate is not trusted.\n"), tag);
+			error_printf_check(_("%s: The certificate is not trusted.\n"), tag);
 		if (status & GNUTLS_CERT_REVOKED)
-			error_printf(_("%s: The certificate has been revoked.\n"), tag);
+			error_printf_check(_("%s: The certificate has been revoked.\n"), tag);
 		if (status & GNUTLS_CERT_SIGNER_NOT_FOUND)
-			error_printf(_("%s: The certificate doesn't have a known issuer.\n"), tag);
+			error_printf_check(_("%s: The certificate doesn't have a known issuer.\n"), tag);
 		if (status & GNUTLS_CERT_SIGNER_NOT_CA)
-			error_printf(_("%s: The certificate signer was not a CA.\n"), tag);
+			error_printf_check(_("%s: The certificate signer was not a CA.\n"), tag);
 		if (status & GNUTLS_CERT_INSECURE_ALGORITHM)
-			error_printf(_("%s: The certificate was signed using an insecure algorithm.\n"), tag);
+			error_printf_check(_("%s: The certificate was signed using an insecure algorithm.\n"), tag);
 		if (status & GNUTLS_CERT_NOT_ACTIVATED)
-			error_printf(_("%s: The certificate is not yet activated.\n"), tag);
+			error_printf_check(_("%s: The certificate is not yet activated.\n"), tag);
 		if (status & GNUTLS_CERT_EXPIRED)
-			error_printf(_("%s: The certificate has expired.\n"), tag);
+			error_printf_check(_("%s: The certificate has expired.\n"), tag);
 #if GNUTLS_VERSION_NUMBER >= 0x030100
 		if (status & GNUTLS_CERT_SIGNATURE_FAILURE)
-			error_printf(_("%s: The certificate signature is invalid.\n"), tag);
+			error_printf_check(_("%s: The certificate signature is invalid.\n"), tag);
 		if (status & GNUTLS_CERT_UNEXPECTED_OWNER)
-			error_printf(_("%s: The certificate's owner does not match hostname '%s'.\n"), tag, hostname);
+			error_printf_check(_("%s: The certificate's owner does not match hostname '%s'.\n"), tag, hostname);
 #endif
 
 		// any other reason
@@ -1022,7 +1083,7 @@ static int verify_certificate_callback(gnutls_session_t session)
 			|GNUTLS_CERT_UNEXPECTED_OWNER
 #endif
 			))
-			error_printf(_("%s: The certificate could not be verified (0x%X).\n"), tag, status);
+			error_printf_check(_("%s: The certificate could not be verified (0x%X).\n"), tag, status);
 
 		goto out;
 	}
@@ -1033,23 +1094,23 @@ static int verify_certificate_callback(gnutls_session_t session)
 	 * be easily extended to work with openpgp keys as well.
 	 */
 	if (gnutls_certificate_type_get(session) != GNUTLS_CRT_X509) {
-		error_printf(_("%s: Certificate must be X.509\n"), tag);
+		error_printf_check(_("%s: Certificate must be X.509\n"), tag);
 		goto out;
 	}
 
 	if (gnutls_x509_crt_init(&cert) != GNUTLS_E_SUCCESS) {
-		error_printf(_("%s: Error initializing X.509 certificate\n"), tag);
+		error_printf_check(_("%s: Error initializing X.509 certificate\n"), tag);
 		goto out;
 	}
 	deinit_cert = 1;
 
 	if (!(cert_list = gnutls_certificate_get_peers(session, &cert_list_size))) {
-		error_printf(_("%s: No certificate was found!\n"), tag);
+		error_printf_check(_("%s: No certificate was found!\n"), tag);
 		goto out;
 	}
 
 	if ((err = gnutls_x509_crt_import(cert, &cert_list[0], GNUTLS_X509_FMT_DER)) != GNUTLS_E_SUCCESS) {
-		error_printf(_("%s: Failed to parse certificate: %s\n"), tag, gnutls_strerror (err));
+		error_printf_check(_("%s: Failed to parse certificate: %s\n"), tag, gnutls_strerror (err));
 		goto out;
 	}
 
@@ -1069,13 +1130,20 @@ static int verify_certificate_callback(gnutls_session_t session)
 //				_get_cert_fingerprint(cert, fingerprint, sizeof(fingerprint)); // calc hexadecimal fingerprint string
 				add_cert_to_ocsp_cache(cert, true);
 				nvalid = 1;
+				skip_server_cert_check = true;
 			}
 #if GNUTLS_VERSION_NUMBER >= 0x030400
-			else if (gnutls_ocsp_status_request_is_checked(session, GNUTLS_OCSP_SR_IS_AVAIL))
-				error_printf(_("WARNING: The certificate's (stapled) OCSP status is invalid\n"));
+			else if (gnutls_ocsp_status_request_is_checked(session, GNUTLS_OCSP_SR_IS_AVAIL)) {
+				error_printf_check(_("WARNING: The certificate's (stapled) OCSP status is invalid\n"));
+				skip_server_cert_check = true;
+			}
 #endif
-			else if (!config.ocsp)
-				error_printf(_("WARNING: The certificate's (stapled) OCSP status has not been sent\n"));
+			else if (!config.ocsp) {
+				debug_printf("OCSP stapling is not supported by '%s'\n", hostname);
+			} else {
+				error_printf_check(_("WARNING: OCSP stapling is not supported by '%s', but OCSP validation has been requested.\n"), hostname);
+				error_printf_check(_("WARNING: This implies a privacy leak: the client sends the certificate serial ID over HTTP to the CA.\n"));
+			}
 #endif
 		} else if (ctx->valid)
 			debug_printf("OCSP: Host '%s' is valid (from cache)\n", hostname);
@@ -1087,7 +1155,7 @@ static int verify_certificate_callback(gnutls_session_t session)
 		gnutls_x509_crt_init(&cert);
 
 		if ((err = gnutls_x509_crt_import(cert, &cert_list[it], GNUTLS_X509_FMT_DER)) != GNUTLS_E_SUCCESS) {
-			error_printf(_("%s: Failed to parse certificate[%u]: %s\n"), tag, it, gnutls_strerror (err));
+			error_printf_check(_("%s: Failed to parse certificate[%u]: %s\n"), tag, it, gnutls_strerror (err));
 			continue;
 		}
 
@@ -1097,55 +1165,55 @@ static int verify_certificate_callback(gnutls_session_t session)
 		cert_verify_hpkp(cert, hostname, session);
 
 #ifdef WITH_OCSP
-		if (config.ocsp && it > nvalid) {
-			char fingerprint[64 * 2 +1];
-			int revoked;
+		if (!config.ocsp || (skip_server_cert_check && it == 0))
+			continue;
 
-			_get_cert_fingerprint(cert, fingerprint, sizeof(fingerprint)); // calc hexadecimal fingerprint string
+		char fingerprint[64 * 2 +1];
+		_get_cert_fingerprint(cert, fingerprint, sizeof(fingerprint)); // calc hexadecimal fingerprint string
 
-			if (wget_ocsp_fingerprint_in_cache(config.ocsp_cert_cache, fingerprint, &revoked)) {
-				// found cert's fingerprint in cache
-				if (revoked) {
-					debug_printf("Certificate[%u] of '%s' has been revoked (cached)\n", it, hostname);
-					nrevoked++;
-				} else {
-					debug_printf("Certificate[%u] of '%s' is valid (cached)\n", it, hostname);
-					nvalid++;
-				}
-				continue;
-			}
-
-			if (deinit_issuer) {
-				gnutls_x509_crt_deinit(issuer);
-				deinit_issuer = 0;
-			}
-			if ((err = gnutls_certificate_get_issuer(credentials, cert, &issuer, 0)) != GNUTLS_E_SUCCESS && it < cert_list_size - 1) {
-				gnutls_x509_crt_init(&issuer);
-				deinit_issuer = 1;
-				if ((err = gnutls_x509_crt_import(issuer, &cert_list[it + 1], GNUTLS_X509_FMT_DER))  != GNUTLS_E_SUCCESS) {
-					debug_printf("Decoding error: %s\n", gnutls_strerror(err));
-					continue;
-				}
-			} else if (err  != GNUTLS_E_SUCCESS) {
-				debug_printf("Cannot find issuer: %s\n", gnutls_strerror(err));
-				continue;
-			}
-
-			ocsp_ok = cert_verify_ocsp(cert, issuer);
-			debug_printf("check_ocsp_response() returned %d\n", ocsp_ok);
-
-			if (ocsp_ok == 1) {
-				debug_printf("Certificate[%u] of '%s' is valid (via OCSP)\n", it, hostname);
-				wget_ocsp_db_add_fingerprint(config.ocsp_cert_cache, fingerprint, time(NULL) + 3600, true); // 1h valid
-				nvalid++;
-			} else if (ocsp_ok == 0) {
-				debug_printf("%s: Certificate[%u] of '%s' has been revoked (via OCSP)\n", tag, it, hostname);
-				wget_ocsp_db_add_fingerprint(config.ocsp_cert_cache, fingerprint, time(NULL) + 3600, false);  // cert has been revoked
+		int revoked;
+		if (wget_ocsp_fingerprint_in_cache(config.ocsp_cert_cache, fingerprint, &revoked)) {
+			// found cert's fingerprint in cache
+			if (revoked) {
+				debug_printf("Certificate[%u] of '%s' has been revoked (cached)\n", it, hostname);
 				nrevoked++;
 			} else {
-				debug_printf("WARNING: OCSP response not available or ignored\n");
-				nignored++;
+				debug_printf("Certificate[%u] of '%s' is valid (cached)\n", it, hostname);
+				nvalid++;
 			}
+			continue;
+		}
+
+		if (deinit_issuer) {
+			gnutls_x509_crt_deinit(issuer);
+			deinit_issuer = 0;
+		}
+		if ((err = gnutls_certificate_get_issuer(credentials, cert, &issuer, 0)) != GNUTLS_E_SUCCESS && it < cert_list_size - 1) {
+			gnutls_x509_crt_init(&issuer);
+			deinit_issuer = 1;
+			if ((err = gnutls_x509_crt_import(issuer, &cert_list[it + 1], GNUTLS_X509_FMT_DER))  != GNUTLS_E_SUCCESS) {
+				debug_printf("Decoding error: %s\n", gnutls_strerror(err));
+				continue;
+			}
+		} else if (err  != GNUTLS_E_SUCCESS) {
+			debug_printf("Cannot find issuer: %s\n", gnutls_strerror(err));
+			continue;
+		}
+
+		ocsp_ok = cert_verify_ocsp(cert, issuer);
+		debug_printf("check_ocsp_response() returned %d\n", ocsp_ok);
+
+		if (ocsp_ok == 1) {
+			debug_printf("Certificate[%u] of '%s' is valid (via OCSP)\n", it, hostname);
+			wget_ocsp_db_add_fingerprint(config.ocsp_cert_cache, fingerprint, time(NULL) + 3600, true); // 1h valid
+			nvalid++;
+		} else if (ocsp_ok == 0) {
+			debug_printf("%s: Certificate[%u] of '%s' has been revoked (via OCSP)\n", tag, it, hostname);
+			wget_ocsp_db_add_fingerprint(config.ocsp_cert_cache, fingerprint, time(NULL) + 3600, false);  // cert has been revoked
+			nrevoked++;
+		} else {
+			debug_printf("WARNING: OCSP response not available or ignored\n");
+			nignored++;
 		}
 #endif
 	}
@@ -1173,7 +1241,7 @@ static int verify_certificate_callback(gnutls_session_t session)
 #endif
 
 	if (!pinning_ok) {
-		error_printf(_("%s: Pubkey pinning mismatch!\n"), tag);
+		error_printf_check(_("%s: Pubkey pinning mismatch!\n"), tag);
 		ret = -1;
 	}
 
@@ -1191,18 +1259,25 @@ out:
 static int init;
 static wget_thread_mutex mutex;
 
-static void __attribute__ ((constructor)) tls_init(void)
-{
-	if (!mutex)
-		wget_thread_mutex_init(&mutex);
-}
-
-static void __attribute__ ((destructor)) tls_exit(void)
+static void tls_exit(void)
 {
 	if (mutex)
 		wget_thread_mutex_destroy(&mutex);
 }
 
+INITIALIZER(tls_init)
+{
+	if (!mutex) {
+		wget_thread_mutex_init(&mutex);
+
+		// Initialize paths while in a thread-safe environment (mostly for _WIN32).
+		wget_ssl_default_cert_dir();
+		wget_ssl_default_ca_bundle_path();
+
+		atexit(tls_exit);
+	}
+}
+
 static int key_type(int type)
 {
 	if (type == WGET_SSL_X509_FMT_DER)
@@ -1236,6 +1311,8 @@ static void set_credentials(gnutls_certificate_credentials_t creds)
 			error_printf(_("No certificates or keys were found\n"));
 	}
 
+	if (config.ca_file && !wget_strcmp(config.ca_file, "system"))
+		config.ca_file = wget_ssl_default_ca_bundle_path();
 	if (config.ca_file) {
 		if (gnutls_certificate_set_x509_trust_file(creds, config.ca_file, key_type(config.ca_type)) <= 0)
 			error_printf(_("No CAs were found in '%s'\n"), config.ca_file);
@@ -1293,20 +1370,23 @@ void wget_ssl_init(void)
 				ncerts = 0;
 
 				if (!strcmp(config.ca_directory, "system"))
-					config.ca_directory = "/etc/ssl/certs";
+					config.ca_directory = wget_ssl_default_cert_dir();
 
 				if ((dir = opendir(config.ca_directory))) {
 					struct dirent *dp;
-					size_t dirlen = strlen(config.ca_directory);
 
 					while ((dp = readdir(dir))) {
 						size_t len = strlen(dp->d_name);
 
 						if (len >= 4 && !wget_strncasecmp_ascii(dp->d_name + len - 4, ".pem", 4)) {
-							struct stat st;
-							char fname[dirlen + 1 + len + 1];
+							char *fname = wget_aprintf("%s/%s", config.ca_directory, dp->d_name);
 
-							wget_snprintf(fname, sizeof(fname), "%s/%s", config.ca_directory, dp->d_name);
+							if (!fname) {
+								error_printf(_("Failed to allocate file name for cert '%s/%s'\n"), config.ca_directory, dp->d_name);
+								continue;
+							}
+
+							struct stat st;
 							if (stat(fname, &st) == 0 && S_ISREG(st.st_mode)) {
 								debug_printf("GnuTLS loading %s\n", fname);
 								if ((rc = gnutls_certificate_set_x509_trust_file(credentials, fname, GNUTLS_X509_FMT_PEM)) <= 0)
@@ -1314,6 +1394,8 @@ void wget_ssl_init(void)
 								else
 									ncerts += rc;
 							}
+
+							xfree(fname);
 						}
 					}
 
@@ -1592,6 +1674,10 @@ int wget_ssl_open(wget_tcp *tcp)
 	if (!tcp)
 		return WGET_E_INVALID;
 
+	struct session_context *ctx = wget_calloc(1, sizeof(struct session_context));
+	if (!ctx)
+		return WGET_E_MEMORY;
+
 	if (!init)
 		wget_ssl_init();
 
@@ -1636,24 +1722,27 @@ int wget_ssl_open(wget_tcp *tcp)
 		gnutls_session_enable_compatibility_mode(session);
 
 	// RFC 6066 SNI Server Name Indication
-	if (hostname)
+	if (hostname) {
 		gnutls_server_name_set(session, GNUTLS_NAME_DNS, hostname, strlen(hostname));
+		debug_printf("SNI %s\n", hostname);
+	}
 	gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, credentials);
 
-	struct session_context *ctx = wget_calloc(1, sizeof(struct session_context));
 	ctx->hostname = wget_strdup(hostname);
+	ctx->port = tcp->remote_port;
 
 #ifdef WITH_OCSP
 	// If we know the cert chain for the hostname being valid at the moment,
 	// we don't ask for OCSP stapling to avoid unneeded IP traffic.
 	// In the unlikely case that the server's certificate chain changed right now,
-	// we fallback to OCSP responder request later.
+	// we fallback to OCSP responder request later (if enabled).
 	if (hostname) {
 		if (!(ctx->valid = wget_ocsp_hostname_is_valid(config.ocsp_host_cache, hostname))) {
 #if GNUTLS_VERSION_NUMBER >= 0x030103
-			if ((rc = gnutls_ocsp_status_request_enable_client(session, NULL, 0, NULL)) == GNUTLS_E_SUCCESS)
+			if ((rc = gnutls_ocsp_status_request_enable_client(session, NULL, 0, NULL)) == GNUTLS_E_SUCCESS) {
+				debug_printf("OCSP stapling requested for %s\n", hostname);
 				ctx->ocsp_stapling = 1;
-			else
+			} else
 				error_printf("GnuTLS: %s\n", gnutls_strerror(rc)); // no translation
 #endif
 		}
